/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef CONSTRAINTDB_HPP_
#define CONSTRAINTDB_HPP_

#include	"Types.hpp"
#include	"Statistics.hpp"
#include	"Settings.hpp"
#include	<vector>


using std::vector;

class ConstraintDB;


const uint32_t INVALID_LEVEL	 	= ~zero32;
const uint8_t VAR_VALUE_MASK		= 3; //11
const uint8_t VAR_VALUE_MASK_INV	= ~VAR_VALUE_MASK; //11
const uint32_t BINARY_MASK 			= 1;
const uint32_t TERNARY_MASK 		= 1 << 1;
const uint32_t SHORT_MASK			= BINARY_MASK | TERNARY_MASK;

/**Helper class to watch particular clauses.
 * Clauses of size <= 3 are stored directly in the watchers list
 * Additionally one literal of the clause is cached for fast detection if the clause is satisfied*/
class WatchedElement{

	//cached literal of the clause, two bits for the type of watched element
	uint32_t literal30_tern_bin;

	//for binary clauses this is not used, for ternary clauses this is the third literal
	//and for longer clauses this is a pointer to the watched clause
	union{Clause*  cls; uint32_t lit_idx;} second;

public:

	//constructor for binary clauses
	WatchedElement(const Literal& lit):literal30_tern_bin((lit.idx() << 2)|BINARY_MASK){second.cls = NULL; }

	//constructor for ternary clauses
	WatchedElement(const Literal& lit1, const Literal& lit2):
		literal30_tern_bin((lit1.idx() << 2)| TERNARY_MASK){assert(lit2 != INVALID_LIT);second.lit_idx = lit2.idx(); }

	//constructor for longer clauses
	WatchedElement(const Literal& lit1, Clause* cls):
	   	literal30_tern_bin(lit1.idx() << 2){second.cls = cls; }

	inline WatchedElement& operator=(const WatchedElement& copy);

	//needed for std::vector operations
	WatchedElement()						 {WatchedElement(INVALID_LIT);}

	WatchedElement(const WatchedElement& copy):
			literal30_tern_bin(copy.literal30_tern_bin),second(copy.second){}

	Literal  getFirstLiteral()			const   {return Literal(literal30_tern_bin >> 2);}
	bool	 isBinary()					const   {return literal30_tern_bin & BINARY_MASK;}
	bool	 isTernary()				const   {return literal30_tern_bin & TERNARY_MASK;}
	bool	 isShort()					const   {return literal30_tern_bin & SHORT_MASK;}
	Clause*  getClause()				const   {assert(!isShort());return second.cls;}
	Literal  getSecondLiteral()			const   {assert(isTernary());return Literal(second.lit_idx);}

	void	 setFirstLiteral(const Literal& lit){
		   literal30_tern_bin = (lit.idx() << 2) | (literal30_tern_bin & SHORT_MASK);}

	void	 setSecondLiteral(const Literal& lit){
		assert(isTernary());
		second.lit_idx = lit.idx();}

	//to find a particular watched clause
	struct ClauseEqualFunctor{
		const Clause* comp;
		ClauseEqualFunctor(Clause* compare):comp(compare){assert(compare != NULL); };
		bool operator() (const WatchedElement& e){return !e.isShort() &&  e.second.cls == comp;}
	};

	//to find a binary clause
	struct BinaryEqualFunctor{
	   	const uint32_t comp;
		   BinaryEqualFunctor(Literal compare):comp((compare.idx() << 2)|BINARY_MASK){};
		   bool operator() (const WatchedElement& e){return e.literal30_tern_bin == comp;}
	};
};

//useful typedefs
typedef vector<WatchedElement> we_vec;
typedef	vector<WatchedElement> ::iterator we_iter;


//since we differ between binary and ternary clauses we also have to differ for the assignment reasons
//They are needed to analyse the reason for conflicting assignments
class AssignInfo{
	//level at which the assignment was made
	uint32_t 	level;

	//states which kind of reason we are dealing with (binary, ternary or any longer clause)
	//for assignments that were caused by ternary clauses seclit30 stores the second literal
	uint32_t 	seclit30_tern_bin;

	//for binary and ternary reasons we keep the literal
	//else a pointer to the clause that became unit under a particular assignment
	union{uint32_t lit; Clause* cls;} reason;

public:
	AssignInfo()									 :level(INVALID_LEVEL),seclit30_tern_bin(0)	{reason.cls = NULL;}
	AssignInfo(const uint32_t lev,Clause* c = NULL)  :level(lev),seclit30_tern_bin(0) 			{reason.cls = c;}
	AssignInfo(const uint32_t lev,const Literal& l)  :level(lev),seclit30_tern_bin(BINARY_MASK) {reason.lit = l.idx();}
	AssignInfo(const uint32_t lev,const Literal& l1,const Literal& l2)
		:level(lev),seclit30_tern_bin((l2.idx() << 2) | TERNARY_MASK) {reason.lit = l1.idx();}

	void		clear()						{ level = INVALID_LEVEL; seclit30_tern_bin = 0; reason.cls = NULL;}
	bool		isBinReason()		const	{ return seclit30_tern_bin & BINARY_MASK;}
	bool		isTernReason()		const	{ return seclit30_tern_bin & TERNARY_MASK;}
	bool		isShortReason()		const	{ return seclit30_tern_bin & SHORT_MASK;}
	bool		isClsReason()		const	{ return !isShortReason();}
	bool		isReason(Clause* cl)const	{ return isClsReason() && cl == reason.cls;}
	bool		noReason()			const	{ return isClsReason() && reason.cls == NULL;}

	Literal		getShortFirstReason()const	{ assert(isShortReason());return Literal(reason.lit);}
	Literal		getShortSecondReason()const	{ assert(isTernReason());return Literal(seclit30_tern_bin >> 2);}
	Clause*		getClsReason()  	const	{ assert(isClsReason());return reason.cls;}
	uint32_t	getLevel()	  		const	{ assert(level != INVALID_LEVEL);return level;}

	void setBinReason(const uint32_t lev,const Literal& l){
		level  = lev; seclit30_tern_bin = BINARY_MASK; reason.lit  = l.idx();}

	void setTernReason(const uint32_t lev,const Literal& l1, const Literal& l2){
			level  = lev; seclit30_tern_bin = (l2.idx() << 2) | TERNARY_MASK; reason.lit  = l1.idx();}

	void setClsReason(const uint32_t lev,Clause* c){
		level = lev;  seclit30_tern_bin = 0; reason.cls  = c;}
};



/**Class to store all information about the formula to be solved. */
class ConstraintDB{

protected:
	//all user settings
	Settings&				settings;

	Statistics&				statistics;

	inline					ConstraintDB(Settings& setting, Statistics& stats);

	inline					~ConstraintDB();



//Variable related Information
private:
	/**An assignment value for each variable and the previous (false) assignment*/
	vector< uint8_t>		assignments;

	/**An assignment level for each variable*/
	vector< AssignInfo>		assign_info;


protected:
	/**Tells the solver to reserve enough space to hold the given number of variables.*/
	void					reserveForVariables(uint32_t sz);

	/**Assign a value to the variable without checking the current value.*/
	inline void				varUncheckedAssign(const Literal& lit,Clause* reason,uint32_t lev);
	inline void				varUncheckedAssign(const Literal& lit,const Literal& reason, uint32_t lev);
	inline void				varUncheckedAssign(const Literal& lit,const Literal& reason1,const Literal& reason2,uint32_t lev);

	/**Clear assignment of a variable*/
	inline void				varClearAssignmentSetPrev(const Variable& v);
	inline void				varClearAssignmentSetPrevFlp(const Variable& v);
	inline void				varClearAssignmentSimple(const Variable& v);

	/**Return the reason for the assignment of a variable.*/
	inline const AssignInfo& varAssignInfo(const Variable& v)	const;

public:
	/**Return the current value of a variable*/
	inline TrBool			varValue(const Variable& v)			const;

	/**Return the previous value of a variable or TR_False if the variable was not assigned before*/
	inline TrBool			varPrevValue(const Variable& v)		const;

	/**Return the level of assignment of a variable. For unassigned variables this is INVALID_LEVEL*/
	inline const uint32_t	varAssignLevel(const Variable& v)	const;



//Literal related Information
private:
	/**For each literal it holds a vector of clauses/literals that are watched*/
	vector<we_vec > watches;

protected:
	inline we_vec&  		litWatchedClauses(const Literal& lit);

	/**Add a clause with size at least 4 to be watched to the specified literal.
	 * @param lit: literal to watch the clause
	 * @param other_watched: the literal of the other watcher in the clause
	 * @param cl: clause to be watched*/
	inline void				litAddWatchedClause(const Literal& lit, const Literal& other_watched, Clause* cl);

public:
	inline bool				litIsTrue(const Literal& l)			const;
	inline bool				litIsNotFalse(const Literal& l)		const;
	inline bool				litIsFalse(const Literal& l)		const;
	inline bool				litIsUnknown(const Literal& l)		const;
	inline void				litPrintWatches(const Literal& l)	const;
	inline const uint32_t	litAssignLevel(const Literal& l)	const;


//Clause related Information
//all arrays store according information to one clause each - indexed by the ID of a clause
protected:
	/**Holds a pointer to all clauses with size > 4.*/
	vector<Clause*>			fix_cls;

	/**Holds a pointer to all learnt clauses with size > 4.*/
	vector<Clause*>			learnt_cls;

private:
	uint32_t				nbr_bin_clses;
	uint32_t				nbr_tern_clses;
	uint64_t				nbr_short_literals;
	uint64_t				nbr_long_learnt_literals;
	uint64_t				nbr_long_fix_literals;

protected:
	cls_act_t				cls_inc;

	void					clsUncreate(Clause* cl,bool unlink = true);

	inline bool				clsLocked(Clause& cls);


protected:
	/**Create a clause from the specified literals (at least 3!!).
	 * @param lits: literals of the clause to be created
	 * @param learnt: indicates if the clause to create is a learnt clause
	 * @return the newly created clause*/
	Clause*					clsCreate(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend, bool learnt = false);

	void					clsCreateBinary(const Literal lit0, const Literal lit1);
	void					clsCreateTernary(const Literal lit0, const Literal lit1, const Literal lit2 );

	inline bool				containsBinary(const Literal lit0, const Literal lit1);

	inline void				clsUnlinkWatched(Clause* cl);
	inline void				clsLinkWatched(Clause* cl);

	inline void				clsBump(Clause* cl);

	void					learntsGC(uint32_t new_sz);

	inline void				clsClearAllStamps();

	inline void				clsShrinkToSize(Clause& cl, uint32_t sz);
	inline void				clsRemoveLiteral(Clause& cl, uint32_t index);

	void					removeSat(vector<Clause*>& cls);


public:
	/**Tell the ConstraintDB instance to reserve enough space to hold the given number of clauses.*/
	inline void				reserveForClauses(uint32_t sz);


//general
public:

	uint32_t numberVariables()		  			const   {return assignments.size();}
	uint64_t numberLiteralsInLongLearntClses()	const	{return nbr_long_learnt_literals;}
	uint64_t numberLiteralsInLongFixClses()		const	{return nbr_long_fix_literals;}
	uint64_t numberLiteralsInLongClses()		const	{return numberLiteralsInLongLearntClses()+numberLiteralsInLongFixClses();}
	uint64_t numberLiteralsInShortClses()		const	{return nbr_short_literals;}
	uint64_t numberLiterals()					const	{return numberLiteralsInLongClses()+numberLiteralsInShortClses();}

	uint32_t numberBinClses()					const	{return nbr_bin_clses;}
	uint32_t numberTernClses()					const	{return nbr_tern_clses;}
	uint32_t numberShortClses()					const	{return numberBinClses() + numberTernClses();}
	uint32_t numberLearntLongClses()			const	{return learnt_cls.size();}
	uint32_t numberOrigLongClses()				const	{return fix_cls.size();}
	uint32_t numberAllLongClses()				const	{return numberOrigLongClses() + numberLearntLongClses();}
	uint32_t numberAllClauses()					const	{return numberAllLongClses() + numberShortClses();}


	bool			clsCanBeOmitted(Clause& cl);

	  /**Check if the given clause is satisfied under the current assignment.*/
	inline bool		checkSatisfied(Clause& cl)								  const;

	void			writeDimacs(OFStream& out, bool skip_sat_cls=false)	const;

	void			printModel();

	//Debug
	bool debugVerifyModel();
};





//Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#define ch_idx(index,vector) assert(index.ID >= 0 && (uint32_t)index.ID < vector .size() )


//Implementation WatchedElement

WatchedElement& WatchedElement::operator=(const WatchedElement& copy){
	literal30_tern_bin  = copy.literal30_tern_bin;
	second			 	= copy.second;
	return *this;
}


//Implementation ConstraintDB
TrBool		ConstraintDB::varValue(const Variable& v)		const   {ch_idx(v,assignments);return TrBool(assignments[v.ID]&VAR_VALUE_MASK);}
TrBool		ConstraintDB::varPrevValue(const Variable& v)	const   {ch_idx(v,assignments);return TrBool((assignments[v.ID]>>2)&VAR_VALUE_MASK);}
const AssignInfo&
			ConstraintDB::varAssignInfo(const Variable& v)	const	{ch_idx(v,assign_info); return assign_info[v.ID];}
bool		ConstraintDB::litIsTrue(const Literal& l)		const   {return varValue(l.toVar()) == l.pol();}
bool		ConstraintDB::litIsNotFalse(const Literal& l)	const   {return varValue(l.toVar()) != l.flpdPolarity();}
bool		ConstraintDB::litIsFalse(const Literal& l)		const   {return varValue(l.toVar()) == l.flpdPolarity();}
bool		ConstraintDB::litIsUnknown(const Literal& l)	const   {return varValue(l.toVar()).isUnknown();}

void		ConstraintDB::varClearAssignmentSetPrev(const Variable& v)	 {
	ch_idx(v,assignments);
	assignments[v.ID] = (assignments[v.ID] << 2) | UNKNOWN_TRBOOL_VAL;}

void		ConstraintDB::varClearAssignmentSetPrevFlp(const Variable& v)  {
	ch_idx(v,assignments);
	assert(TrBool(assignments[v.ID]&VAR_VALUE_MASK).isKnown());
	assignments[v.ID] = ((assignments[v.ID] ^ TRBOOL_FLIP_MASK) << 2) | UNKNOWN_TRBOOL_VAL;}

void		ConstraintDB::varClearAssignmentSimple(const Variable& v)	 {
	ch_idx(v,assignments);
	assignments[v.ID] = (assignments[v.ID] & VAR_VALUE_MASK_INV) | UNKNOWN_TRBOOL_VAL;}


const uint32_t  ConstraintDB::varAssignLevel(const Variable& v) const
			{ch_idx(v,assign_info); return assign_info[v.ID].getLevel();}
const uint32_t  ConstraintDB::litAssignLevel(const Literal& l)  const
			{ch_idx(l.toVar(),assign_info); return assign_info[l.toVar().ID].getLevel();}


void		ConstraintDB::clsBump(Clause* cl){
	assert(cl->isLearnt());
	if((cl->activity += cls_inc) > settings.cls_max_act ) {
		statistics.nbr_expensive_cls_bmp += 1;
		const cls_act_t::denominator_t w = settings.cls_min_inc.getLogOfDenominator();
		cls_inc.weaken(w);
		uint i = learnt_cls.size();
		while(i--) learnt_cls[i]->activity.weaken(w);
	}
}

void		ConstraintDB::clsShrinkToSize(Clause& cl, uint32_t sz){
	assert(sz < cl.size());
	if(cl.isLearnt())	nbr_long_learnt_literals 	-= (cl.size() - sz);
	else 				nbr_long_fix_literals		-= (cl.size() - sz);
	cl.shrinkToSize(sz);
}

void	ConstraintDB::clsRemoveLiteral(Clause& cl, uint32_t index){
	assert(index < cl.size());
	statistics.nbr_removed_cls_lits += 1;
	cl.removeAt(index);
	if(cl.isLearnt())	--nbr_long_learnt_literals;
	else				--nbr_long_fix_literals;
}


bool ConstraintDB::containsBinary(const Literal lit0, const Literal lit1){
	bool contained = find(litWatchedClauses(lit0),WatchedElement::BinaryEqualFunctor(lit1));
	assert(contained == find(litWatchedClauses(lit1),WatchedElement::BinaryEqualFunctor(lit0)));
	return contained;
}


void	ConstraintDB::clsLinkWatched(Clause* cl){
	const Literal lit0 = (*cl)[0];
	const Literal lit1 = (*cl)[1];

	assert(!find(litWatchedClauses(lit0),WatchedElement::ClauseEqualFunctor(cl)));
	assert(!find(litWatchedClauses(lit1),WatchedElement::ClauseEqualFunctor(cl)));

	litAddWatchedClause(lit0, lit1,cl);
	litAddWatchedClause(lit1, lit0,cl);
}


void	ConstraintDB::clsUnlinkWatched(Clause* cl){
	assert(cl != NULL);
	WatchedElement::ClauseEqualFunctor cl_equal(cl);

	assert(find(litWatchedClauses((*cl)[0]),cl_equal));
	stable_remove_one(litWatchedClauses((*cl)[0]),cl_equal);

	assert(find(litWatchedClauses((*cl)[1]),cl_equal));
	stable_remove_one(litWatchedClauses((*cl)[1]),cl_equal);
}





template<class C>
void free_all(vector<C>& vec){
	uint32_t i(vec.size());
	while(i--){
	   assert(vec[i] != NULL);
	   free(vec[i]);
	   vec[i] =NULL;
	}
}

ConstraintDB::~ConstraintDB(){
	free_all(fix_cls);
	free_all(learnt_cls);
};

ConstraintDB::ConstraintDB(Settings& sets, Statistics& stats):
	settings(sets),
	statistics(stats),
	nbr_bin_clses(0),nbr_tern_clses(0),nbr_short_literals(0),
	nbr_long_learnt_literals(0),nbr_long_fix_literals(0),
	cls_inc(sets.cls_min_inc){ }




void		ConstraintDB::varUncheckedAssign(const Literal& lit, Clause* reason, uint32_t lev){
	uint32_t		 var_idx	= lit.toVar().idx();
	assignments	 [var_idx]   = (VAR_VALUE_MASK_INV & assignments[var_idx]) | lit.pol().toInt();
	assign_info	 [var_idx].setClsReason(lev,reason);
	assert(assign_info[var_idx].getLevel() == lev);
	assert(assign_info[var_idx].getClsReason() == reason);
}

void		ConstraintDB::varUncheckedAssign(const Literal& lit, const Literal& reason, uint32_t lev){
	uint32_t		 var_idx	= lit.toVar().idx();
	assignments	 [var_idx]   = (VAR_VALUE_MASK_INV & assignments[var_idx]) | lit.pol().toInt();
	assign_info	 [var_idx].setBinReason(lev,reason);
	assert(assign_info[var_idx].getLevel() == lev);
	assert(assign_info[var_idx].isBinReason() &&  assign_info[var_idx].getShortFirstReason() == reason);
}

void		ConstraintDB::varUncheckedAssign(const Literal& lit, const Literal& reason1, const Literal& reason2, uint32_t lev){
	uint32_t		 var_idx	= lit.toVar().idx();
	assignments	 [var_idx]   = (VAR_VALUE_MASK_INV & assignments[var_idx]) | lit.pol().toInt();
	assign_info	 [var_idx].setTernReason(lev,reason1,reason2);
	assert(assign_info[var_idx].getLevel() == lev);
	assert(assign_info[var_idx].isTernReason() &&
		assign_info[var_idx].getShortFirstReason() == reason1 &&  assign_info[var_idx].getShortSecondReason() == reason2);
}



void	ConstraintDB::reserveForClauses(uint32_t need_sz){
	if(need_sz > fix_cls.size()){fix_cls.reserve(need_sz);}
}


void	ConstraintDB::litAddWatchedClause(const Literal& lit, const Literal& other_watched,Clause* cl){
	ch_idx(lit,watches); assert(cl != NULL && cl->size() > 2);
	assert((*cl)[0] == lit && (*cl)[1] == other_watched || (*cl)[1] == lit && (*cl)[0] == other_watched);
	watches[lit.idx()].push_back(WatchedElement(other_watched,cl));
}

void	ConstraintDB::clsClearAllStamps(){
	for(uint32_t i = fix_cls.size();i--;){fix_cls[i]->clearStamp();}
	for(uint32_t i = learnt_cls.size();i--;){learnt_cls[i]->clearStamp();}
}


we_vec& ConstraintDB::litWatchedClauses(const Literal& lit)
	{ch_idx(lit,watches); return watches[lit.ID];}

bool ConstraintDB::clsLocked(Clause& cls){
	return assign_info[cls[0].toVar().ID].isReason(&cls) ? litIsTrue(cls[0])
		   : (assign_info[cls[1].toVar().ID].isReason(&cls) && litIsTrue(cls[1]));
}



#undef ch_idx

const char*  constraintdb_cpp_version();
#define constraintdb_hpp_version "ConstraintDB.hpp:   $Revision: 1.12 $"




#endif /*CONSTRAINTDB_HPP_*/
