/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de

Many settings and ideas on how to implement efficient data structures and solving are
borrowed from Minisat 20
 */

#include "Statistics.hpp"
#include "BasicSolver.hpp"
#include <signal.h>

#include "Types.hpp"
#include "IFStream.hpp"




/**
 * Some checks
 */

bool check_defines(){

    return
    //sizeof(void*)           ==      32/8                &&
    sizeof(int32_t)         ==      32/8                &&
    sizeof(int64_t)         ==      64/8                &&
    sizeof(uint32_t)        ==      32/8                &&
    sizeof(uint64_t)        ==      64/8                &&
    sizeof(double)          ==      64/8                &&
    sizeof(float)           ==      32/8                &&
    sizeof(Literal)         ==      sizeof(uint32_t)    &&
    sizeof(Variable)        ==      sizeof(uint32_t)    &&
    sizeof(float)           ==      sizeof(uint32_t)    &&
    ((int)false)            ==      0                   &&
    ((int)true)             ==      1;
};



Settings    settings;
Statistics  statistics(settings);


static void SIGQUIT_handler(int signum){

    OFStream out(stderr);
    out.writeNewLine();
    out.setBackgroundPattern(   "|                                                                                            |\n");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
    out.writeCenteredToBG("!!! INFO !!!");
    statistics.printFinalStatistics(out);
    out.writeCenteredToBG("!!! INFO !!!");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
};

static void SIGINT_handler(int signum) {
    OFStream out(stderr);
    out.writeNewLine();
    out.setBackgroundPattern(   "|                                                                                            |\n");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
    out.writeCenteredToBG("!!! KILLED !!!");
    statistics.printFinalStatistics(out);
    out.writeCenteredToBG("!!! KILLED !!!");
    out.writeString(            "|--------------------------------------------------------------------------------------------");
    exit(1);
};


bool parseDIMACS(IFStream& in, BasicSolver& solver) {
    vector<Literal> parsed_lits;
    int32_t  parsed_lit;

    while(in.hasNext()){
        if(in.currEquals('c')) {in.skipLine();}
        else if(in.currEquals('p')){
            if (in == "p cnf"){
                solver.reserveForVariables(in.getNextInt());
                solver.reserveForClauses(in.getNextInt());
                in.skipLine();
            }else{ fprintf(stderr,"PARSE ERROR in file %s \n",settings.cnf_file_name), exit(EXIT_FAILURE); }
        }else{
            while((parsed_lit = in.getNextInt()) != 0){
            	parsed_lits.push_back(Literal::DIMACS2Literal(parsed_lit));
            }
            if(!solver.addClause(parsed_lits)){
            	fprintf(stderr,"c Formula contains an empty clause!! - is unsatisfiable!");
                solver.setInvalid(); return false;
            }
            parsed_lits.clear();
        }
    }
    return true;
}



int main(int argc, char** argv){
	assert((fprintf(stderr,"Asserts will be evaluated\n"),true));
    assert(check_defines());

    BasicSolver solver(settings,statistics);

    if(!settings.read(argc,argv)){
        settings.printUsage();
        return EXIT_FAILURE;
    }

    IFStream in(settings.cnf_file_name);
    if(in.isBad()){
    	fprintf(stderr,"c Could not open CNF file %s ", settings.cnf_file_name);
        return EXIT_FAILURE;
    }



    signal(SIGINT,SIGINT_handler);
    signal(SIGQUIT,SIGQUIT_handler);



    if(settings.console_out != NULL){
    	settings.console_out->writeNewLine();
    	settings.console_out->writeString(
    	    "|############################################################################################|\n");
    	settings.console_out->setBackgroundPattern(
    	    "|                                                                                            |\n");
    }

    parseDIMACS(in,solver);
    in.close();

    statistics.parse_timestmp = statistics.cpuUserTime() ;

    char* result_txt = (char*)"UNKNOWN";

    TrBool result = solver.isSAT( (double)settings.timeout);

    if(result.isTrue()){
    	result_txt = (char*)"SAT";
    	//printf("s SATISFIABLE\n");
    	//solver.printModel();

    }else if(result.isFalse()){
    	result_txt =(char*)"UNSAT";
    	//printf("s UNSATISFIABLE\n");

    }else{
    	//printf("s UNKNOWN\n");
    }

    if(settings.csv_file_name != NULL){
        OFStream out(settings.csv_file_name,false);
        statistics.printCSV(out,&solver,result_txt);
    }

    if(settings.gnuplot_file_name != NULL){
        OFStream out(settings.gnuplot_file_name,false);
        statistics.printGnuPlot(out,&solver,result_txt);
    }


    if(settings.console_out != NULL){
    	settings.console_out->writeString(
    	    "|Instance %s is %s", basename(settings.cnf_file_name),result_txt );
    	statistics.printFinalStatistics(*settings.console_out,&solver);
    	settings.console_out->writeString(
    	    "|############################################################################################|\n");

    }


#ifdef NDEBUG
    exit(0);
#endif

    return 0;
}




const char*  sapperlot_cpp_version(){return "SApperloT.cpp:      $Revision: 1.8 $";}


